from __future__ import annotations

import os

from pluggy import HookimplMarker, PluginManager

from file_keeper.core import storage, upload

from . import spec

hookimpl = HookimplMarker(spec.name)
plugin = PluginManager(spec.name)
plugin.add_hookspecs(spec)


def setup():
    plugin.load_setuptools_entrypoints(spec.name)

    for name in os.getenv("FILE_KEEPER_DISABLED_EXTENSIONS", "").split():
        undesired = plugin.get_plugin(name)
        if plugin.is_registered(undesired):
            plugin.unregister(undesired)

    plugin.hook.register_location_transformers(registry=storage.location_transformers)
    plugin.hook.register_upload_factories(registry=upload.upload_factories)
    plugin.hook.register_adapters(registry=storage.adapters)


if not os.getenv("FILE_KEEPER_NO_SETUP"):
    setup()



from __future__ import annotations

from typing import TYPE_CHECKING

from pluggy import HookspecMarker

if TYPE_CHECKING:
    from file_keeper.core import storage, upload
    from file_keeper.core.registry import Registry


name = "file_keeper_ext"


hookspec = HookspecMarker(name)


@hookspec
def register_adapters(registry: Registry[type[storage.Storage]]): ...


@hookspec
def register_upload_factories(registry: Registry[upload.UploadFactory, type]): ...


@hookspec
def register_location_transformers(registry: Registry[storage.LocationTransformer]): ...



from __future__ import annotations

import contextlib
import io
import mimetypes
import os
import tempfile
import uuid
from datetime import datetime
from typing import Any, cast

import magic
import pytz
from pluggy import HookimplMarker

from file_keeper import Registry, Storage, Upload, ext
from file_keeper.core.storage import LocationTransformer
from file_keeper.core.upload import UploadFactory

from . import adapters

hookimpl = HookimplMarker("file-keeper-ext")
SAMPLE_SIZE = 1024 * 2


@ext.hookimpl
def register_location_transformers(registry: Registry[LocationTransformer]):
    registry.register("safe_relative_path", safe_relative_path_transformer)
    registry.register("uuid", uuid_transformer)
    registry.register("uuid_prefix", uuid_prefix_transformer)
    registry.register("uuid_with_extension", uuid_with_extension_transformer)
    registry.register("datetime_prefix", datetime_prefix_transformer)
    registry.register("datetime_with_extension", datetime_with_extension_transformer)


def safe_relative_path_transformer(location: str, extras: dict[str, Any]) -> str:
    return os.path.normpath(location).lstrip("./")


def uuid_transformer(location: str, extras: dict[str, Any]) -> str:
    return str(uuid.uuid4())


def uuid_prefix_transformer(location: str, extras: dict[str, Any]) -> str:
    return str(uuid.uuid4()) + location


def uuid_with_extension_transformer(location: str, extras: dict[str, Any]) -> str:
    ext = os.path.splitext(location)[1]
    return str(uuid.uuid4()) + ext


def datetime_prefix_transformer(location: str, extras: dict[str, Any]) -> str:
    return datetime.now(pytz.utc).isoformat() + location


def datetime_with_extension_transformer(location: str, extras: dict[str, Any]) -> str:
    ext = os.path.splitext(location)[1]
    return datetime.now(pytz.utc).isoformat() + ext


@ext.hookimpl
def register_upload_factories(registry: Registry[UploadFactory, type]):
    registry.register(tempfile.SpooledTemporaryFile, tempfile_into_upload)
    registry.register(io.TextIOWrapper, textiowrapper_into_upload)


with contextlib.suppress(ImportError):  # pragma: no cover
    import cgi

    @ext.hookimpl(specname="register_upload_factories")
    def _(registry: Registry[UploadFactory, type]):
        registry.register(cgi.FieldStorage, cgi_field_storage_into_upload)

    def cgi_field_storage_into_upload(value: cgi.FieldStorage):
        if not value.filename or not value.file:
            return None

        mime, _encoding = mimetypes.guess_type(value.filename)
        if not mime:
            mime = magic.from_buffer(value.file.read(SAMPLE_SIZE), True)
            _ = value.file.seek(0)

        _ = value.file.seek(0, 2)
        size = value.file.tell()
        _ = value.file.seek(0)

        return Upload(
            value.file,
            value.filename,
            size,
            mime,
        )


with contextlib.suppress(ImportError):  # pragma: no cover
    from werkzeug.datastructures import FileStorage

    @ext.hookimpl(specname="register_upload_factories")
    def _(registry: Registry[UploadFactory, type]):
        registry.register(FileStorage, werkzeug_file_storage_into_upload)

    def werkzeug_file_storage_into_upload(value: FileStorage):
        name: str = value.filename or value.name or ""
        if value.content_length:
            size = value.content_length
        else:
            _ = value.stream.seek(0, 2)
            size = value.stream.tell()
            _ = value.stream.seek(0)

        mime = magic.from_buffer(value.stream.read(SAMPLE_SIZE), True)
        _ = value.stream.seek(0)

        return Upload(value.stream, name, size, mime)


def tempfile_into_upload(value: tempfile.SpooledTemporaryFile[bytes]):
    mime = magic.from_buffer(value.read(SAMPLE_SIZE), True)
    _ = value.seek(0, 2)
    size = value.tell()
    _ = value.seek(0)

    return Upload(value, value.name or "", size, mime)


def textiowrapper_into_upload(value: io.TextIOWrapper):
    return cast(io.BufferedReader, value.buffer)


@ext.hookimpl
def register_adapters(registry: Registry[type[Storage]]):
    registry.register("file_keeper:fs", adapters.FsStorage)

    if adapters.RedisStorage:
        registry.register("file_keeper:redis", adapters.RedisStorage)

    if adapters.OpenDalStorage:
        registry.register("file_keeper:opendal", adapters.OpenDalStorage)

    if adapters.LibCloudStorage:
        registry.register("file_keeper:libcloud", adapters.LibCloudStorage)

    if adapters.GoogleCloudStorage:
        registry.register("file_keeper:gcs", adapters.GoogleCloudStorage)

    if adapters.S3Storage:
        registry.register("file_keeper:s3", adapters.S3Storage)

    if adapters.FilebinStorage:
        registry.register("file_keeper:filebin", adapters.FilebinStorage)

    if adapters.SqlAlchemyStorage:
        registry.register("file_keeper:sqlalchemy", adapters.SqlAlchemyStorage)



"""Exception definitions for the extension.

Hierarchy:

* Exception
    * FilesError
        * StorageError
            * UnknownAdapterError
            * UnknownStorageError
            * UnsupportedOperationError
            * PermissionError
            * LocationError
                * MissingFileError
                * ExistingFileError
            * ExtrasError
                * MissingExtrasError
            * InvalidStorageConfigurationError
                * MissingStorageConfigurationError
            * UploadError
                * WrongUploadTypeError
                * LocationTransformerError
                * ContentError
                * LargeUploadError
                    * UploadOutOfBoundError
                * UploadMismatchError
                    * UploadTypeMismatchError
                    * UploadHashMismatchError
                    * UploadSizeMismatchError

"""

from __future__ import annotations

from typing import Any

Storage = Any


class FilesError(Exception):
    """Base error for catch-all scenario."""


class StorageError(FilesError):
    """Error related to storage."""


class UnknownStorageError(StorageError):
    """Storage with the given name is not configured."""

    def __init__(self, storage: str):
        self.storage = storage

    def __str__(self):
        return f"Storage {self.storage} is not configured"


class UnknownAdapterError(StorageError):
    """Specified storage adapter is not registered."""

    def __init__(self, adapter: str):
        self.adapter = adapter

    def __str__(self):
        return f"Storage adapter {self.adapter} is not registered"


class UnsupportedOperationError(StorageError):
    """Requested operation is not supported by storage."""

    def __init__(self, operation: str, storage: Storage):
        self.operation = operation
        self.storage = storage

    def __str__(self):
        return f"Operation {self.operation} is not supported by {self.storage} storage"


class InvalidStorageConfigurationError(StorageError):
    """Storage cannot be initialized with given configuration."""

    def __init__(self, adapter_or_storage: type | str, problem: str):
        self.adapter_or_storage = adapter_or_storage
        self.problem = problem

    def __str__(self):
        if isinstance(self.adapter_or_storage, str):
            what = f"storage adapter {self.adapter_or_storage}"
        else:
            what = f"storage adapter {self.adapter_or_storage.__name__}"

        return f"Cannot initialize {what}" + f" due to following error: {self.problem}"


class PermissionError(StorageError):
    """Storage client does not have required permissions."""

    def __init__(self, storage: Storage, operation: str, problem: str):
        self.storage = storage
        self.operation = operation
        self.problem = problem

    def __str__(self):
        msg = "Storage {} is not allowed to perform {} operation: {}"
        return msg.format(
            self.storage,
            self.operation,
            self.problem,
        )


class MissingStorageConfigurationError(InvalidStorageConfigurationError):
    """Storage cannot be initialized due to missing option."""

    def __init__(self, adapter_or_storage: type | str, option: str):
        super().__init__(
            adapter_or_storage,
            f"{option} option is required",
        )


class LocationError(StorageError):
    """Storage cannot use given location."""

    def __init__(self, storage: Storage, location: str):
        self.storage = storage
        self.location = location

    def __str__(self):
        return f"{self.storage} cannot use location {self.location}"


class MissingFileError(LocationError):
    """File does not exist."""

    def __str__(self):
        return f"File {self.location} does not exist inside storage {self.storage}"


class ExistingFileError(LocationError):
    """File already exists."""

    def __str__(self):
        return f"File {self.location} already exists inside storage {self.storage}"


class UploadError(StorageError):
    """Error related to file upload process."""


class LargeUploadError(UploadError):
    """Storage cannot be initialized due to missing option."""

    def __init__(self, actual_size: int, max_size: int):
        self.actual_size = actual_size
        self.max_size = max_size

    def __str__(self):
        return (
            f"Upload size {self.actual_size} surpasses"
            + f" max allowed size {self.max_size}"
        )


class UploadOutOfBoundError(LargeUploadError):
    """Multipart upload exceeds expected size."""

    def __str__(self):
        return (
            f"Upload size {self.actual_size} exceeds"
            + f" expected size {self.max_size}"
        )


class UploadMismatchError(UploadError):
    """Expected value of file attribute doesn't match the actual value."""

    value_formatter = str

    def __init__(self, attribute: str, actual: Any, expected: Any):
        self.attribute = attribute
        self.actual = actual
        self.expected = expected

    def __str__(self):
        actual = self.value_formatter(self.actual)
        expected = self.value_formatter(self.expected)
        return (
            f"Actual value of {self.attribute}({actual}) does not"
            + f" match expected value({expected})"
        )


class UploadTypeMismatchError(UploadMismatchError):
    """Expected value of content type doesn't match the actual value."""

    def __init__(self, actual: Any, expected: Any):
        super().__init__("content type", actual, expected)


class UploadHashMismatchError(UploadMismatchError):
    """Expected value of hash match the actual value."""

    def __init__(self, actual: Any, expected: Any):
        super().__init__("content hash", actual, expected)


class UploadSizeMismatchError(UploadMismatchError):
    """Expected value of upload size doesn't match the actual value."""

    def __init__(self, actual: Any, expected: Any):
        super().__init__("upload size", actual, expected)


class WrongUploadTypeError(UploadError):
    """Storage does not support given MIMEType."""

    def __init__(self, content_type: str):
        self.content_type = content_type

    def __str__(self):
        return f"Type {self.content_type} is not supported by storage"


class LocationTransformerError(UploadError):
    """Undefined location transformer."""

    def __init__(self, transformer: str):
        self.transformer = transformer

    def __str__(self):
        return f"Unknown location transformer {self.transformer}"


class ExtrasError(StorageError):
    """Wrong extras passed during upload."""

    def __init__(self, extras: Any):
        self.extras = extras

    def __str__(self):
        return f"Wrong extras: {self.extras}"


class MissingExtrasError(ExtrasError):
    """Wrong extras passed to storage method."""

    def __init__(self, key: Any):
        self.key = key

    def __str__(self):
        return f"Key {self.key} is missing from extras"


class ContentError(UploadError):
    """Storage cannot accept uploaded content."""

    def __init__(self, storage: Storage, msg: str):
        self.storage = storage
        self.msg = msg

    def __str__(self):
        return f"{self.storage} rejected upload: {self.msg}"



from __future__ import annotations

from typing import Callable, Generic, Hashable

from typing_extensions import TypeVar

K = TypeVar("K", default=str, bound=Hashable)
V = TypeVar("V")


class Registry(Generic[V, K]):
    """Mutable collection of objects.

    Example:
    >>> col = Registry()
    >>>
    >>> col.register("one", 1)
    >>> assert col.get("one") == 1
    >>>
    >>> col.reset()
    >>> assert col.get("one") is None
    """

    def __init__(
        self,
        members: dict[K, V] | None = None,
        collector: Callable[[], dict[K, V]] | None = None,
    ):
        if members is None:
            members = {}
        self.members = members
        self.collector = collector

    def __len__(self):
        return len(self.members)

    def __bool__(self):
        return bool(self.members)

    def __iter__(self):
        return iter(self.members)

    def __getitem__(self, key: K):
        return self.members[key]

    def __contains__(self, item: K):
        return item in self.members

    def collect(self):
        if self.collector:
            self.members.update(self.collector())

    def reset(self):
        """Remove all members from registry."""
        self.members.clear()

    def register(self, key: K, member: V):
        """Add a member to registry."""
        self.members[key] = member

    def get(self, key: K) -> V | None:
        """Get the optional member from registry."""
        return self.members.get(key)

    def pop(self, key: K) -> V | None:
        """Remove the member from registry."""
        return self.members.pop(key, None)

    def decorated(self, key: K):
        def decorator(value: V):
            self.register(key, value)
            return value

        return decorator



from __future__ import annotations

import dataclasses
from typing import Any, ClassVar, Iterable

import sqlalchemy as sa

import file_keeper as fk


@dataclasses.dataclass()
class Settings(fk.Settings):
    db_url: str = ""
    table_name: str = ""
    location_column: str = ""
    content_column: str = ""

    engine: sa.Engine = None  # type: ignore
    table: Any = None
    location: Any = None
    content: Any = None

    _required_options: ClassVar[list[str]] = [
        "db_url",
        "table",
        "location_column",
        "content_column",
    ]

    def __post_init__(self, **kwargs: Any):
        super().__post_init__(**kwargs)

        self.engine = sa.create_engine(self.db_url)
        self.location = sa.column(self.location_column)
        self.content = sa.column(self.content_column)
        self.table = sa.table(
            self.table_name,
            self.location,
            self.content,
        )


class Reader(fk.Reader):
    storage: SqlAlchemyStorage
    capabilities = fk.Capability.STREAM

    def stream(self, data: fk.FileData, extras: dict[str, Any]) -> Iterable[bytes]:
        stmt = (
            sa.select(self.storage.settings.content)
            .select_from(self.storage.settings.table)
            .where(self.storage.settings.location == data.location)
        )

        with self.storage.settings.engine.connect() as conn:
            row = conn.execute(stmt).fetchone()

        if row is None:
            raise fk.exc.MissingFileError(self, data.location)

        return row


class Uploader(fk.Uploader):
    storage: SqlAlchemyStorage
    capabilities = fk.Capability.CREATE

    def upload(
        self,
        location: fk.types.Location,
        upload: fk.Upload,
        extras: dict[str, Any],
    ) -> fk.FileData:
        reader = upload.hashing_reader()

        values: dict[Any, Any] = {
            self.storage.settings.location_column: location,
            self.storage.settings.content_column: reader.read(),
        }
        stmt = sa.insert(self.storage.settings.table).values(values)

        with self.storage.settings.engine.connect() as conn:
            conn.execute(stmt)

        return fk.FileData(
            location,
            upload.size,
            upload.content_type,
            reader.get_hash(),
        )


class Manager(fk.Manager):
    storage: SqlAlchemyStorage
    capabilities = fk.Capability.SCAN | fk.Capability.REMOVE

    def scan(self, extras: dict[str, Any]) -> Iterable[str]:
        stmt = sa.select(self.storage.settings.location).select_from(
            self.storage.settings.table
        )
        with self.storage.settings.engine.connect() as conn:
            for row in conn.execute(stmt):
                yield row[0]

    def remove(
        self,
        data: fk.FileData | fk.MultipartData,
        extras: dict[str, Any],
    ) -> bool:
        stmt = sa.delete(self.storage.settings.table).where(
            self.storage.settings.location == data.location,
        )
        with self.storage.settings.engine.connect() as conn:
            conn.execute(stmt)
        return True


class SqlAlchemyStorage(fk.Storage):
    hidden = True
    settings: Settings  # type: ignore
    SettingsFactory = Settings
    UploaderFactory = Uploader
    ManagerFactory = Manager
    ReaderFactory = Reader



"""Base abstract functionality of the extentsion.

All classes required for specific storage implementations are defined
here. Some utilities, like `make_storage` are also added to this module instead
of `utils` to avoid import cycles.

This module relies only on types, exceptions and utils to prevent import
cycles.

"""

from __future__ import annotations

import dataclasses
import logging
from typing import Any, Callable, ClassVar, Iterable, cast

from typing_extensions import Concatenate, ParamSpec, TypeAlias, TypeVar

from . import data, exceptions, types, utils
from .upload import make_upload, Upload
from .registry import Registry

P = ParamSpec("P")
T = TypeVar("T")


log = logging.getLogger(__name__)

LocationTransformer: TypeAlias = Callable[[str, "dict[str, Any]"], str]

adapters = Registry["type[Storage]"]()
location_transformers = Registry[LocationTransformer]()


def requires_capability(capability: utils.Capability):
    def decorator(func: Callable[Concatenate[Storage, P], T]):
        def method(self: Storage, *args: P.args, **kwargs: P.kwargs) -> T:
            if not self.supports(capability):
                raise exceptions.UnsupportedOperationError(str(capability.name), self)
            return func(self, *args, **kwargs)

        return method

    return decorator


class StorageService:
    """Base class for services used by storage.

    StorageService.capabilities reflect all operations provided by the
    service.

    >>> class Uploader(StorageService):
    >>>     capabilities = Capability.CREATE
    """

    capabilities = utils.Capability.NONE

    def __init__(self, storage: Storage):
        self.storage = storage


class Uploader(StorageService):
    """Service responsible for writing data into a storage.

    `Storage` internally calls methods of this service. For example,
    `Storage.upload(location, upload, **kwargs)` results in
    `Uploader.upload(location, upload, kwargs)`.

    Example:
        ```python
        class MyUploader(Uploader):
            def upload(
                self, location: Location, upload: Upload, extras: dict[str, Any]
            ) -> FileData:
                reader = upload.hashing_reader()

                with open(location, "wb") as dest:
                    dest.write(reader.read())

                return FileData(
                    location, upload.size,
                    upload.content_type,
                    reader.get_hash()
                )
        ```
    """

    def upload(
        self,
        location: types.Location,
        upload: Upload,
        extras: dict[str, Any],
    ) -> data.FileData:
        """Upload file using single stream."""
        raise NotImplementedError

    def multipart_start(
        self,
        location: types.Location,
        data: data.MultipartData,
        extras: dict[str, Any],
    ) -> data.MultipartData:
        """Prepare everything for multipart(resumable) upload."""
        raise NotImplementedError

    def multipart_refresh(
        self,
        data: data.MultipartData,
        extras: dict[str, Any],
    ) -> data.MultipartData:
        """Show details of the incomplete upload."""
        raise NotImplementedError

    def multipart_update(
        self,
        data: data.MultipartData,
        extras: dict[str, Any],
    ) -> data.MultipartData:
        """Add data to the incomplete upload."""
        raise NotImplementedError

    def multipart_complete(
        self,
        data: data.MultipartData,
        extras: dict[str, Any],
    ) -> data.FileData:
        """Verify file integrity and finalize incomplete upload."""
        raise NotImplementedError


class Manager(StorageService):
    """Service responsible for maintenance file operations.

    `Storage` internally calls methods of this service. For example,
    `Storage.remove(data, **kwargs)` results in `Manager.remove(data, kwargs)`.

    Example:
        ```python
        class MyManager(Manager):
            def remove(
                self, data: FileData|MultipartData, extras: dict[str, Any]
            ) -> bool:
                os.remove(data.location)
                return True
        ```
    """

    def remove(
        self, data: data.FileData | data.MultipartData, extras: dict[str, Any]
    ) -> bool:
        """Remove file from the storage."""
        raise NotImplementedError

    def exists(self, data: data.FileData, extras: dict[str, Any]) -> bool:
        """Check if file exists in the storage."""
        raise NotImplementedError

    def compose(
        self,
        location: types.Location,
        datas: Iterable[data.FileData],
        extras: dict[str, Any],
    ) -> data.FileData:
        """Combine multipe file inside the storage into a new one."""
        raise NotImplementedError

    def append(
        self,
        data: data.FileData,
        upload: Upload,
        extras: dict[str, Any],
    ) -> data.FileData:
        """Append content to existing file."""
        raise NotImplementedError

    def copy(
        self,
        location: types.Location,
        data: data.FileData,
        extras: dict[str, Any],
    ) -> data.FileData:
        """Copy file inside the storage."""
        raise NotImplementedError

    def move(
        self,
        location: types.Location,
        data: data.FileData,
        extras: dict[str, Any],
    ) -> data.FileData:
        """Move file to a different location inside the storage."""
        raise NotImplementedError

    def scan(self, extras: dict[str, Any]) -> Iterable[str]:
        """List all locations(filenames) in storage."""
        raise NotImplementedError

    def analyze(
        self,
        location: types.Location,
        extras: dict[str, Any],
    ) -> data.FileData:
        """Return all details about filename."""
        raise NotImplementedError


class Reader(StorageService):
    """Service responsible for reading data from the storage.

    `Storage` internally calls methods of this service. For example,
    `Storage.stream(data, **kwargs)` results in `Reader.stream(data, kwargs)`.

    Example:
        ```python
        class MyReader(Reader):
            def stream(
                self, data: FileData, extras: dict[str, Any]
            ) -> Iterable[bytes]:
                return open(data.location, "rb")
        ```
    """

    def stream(self, data: data.FileData, extras: dict[str, Any]) -> Iterable[bytes]:
        """Return byte-stream of the file content."""
        raise NotImplementedError

    def content(self, data: data.FileData, extras: dict[str, Any]) -> bytes:
        """Return file content as a single byte object."""
        return b"".join(self.stream(data, extras))

    def range(
        self,
        data: data.FileData,
        start: int,
        end: int | None,
        extras: dict[str, Any],
    ) -> Iterable[bytes]:
        """Return slice of the file content."""
        raise NotImplementedError

    def permanent_link(self, data: data.FileData, extras: dict[str, Any]) -> str:
        """Return permanent download link."""
        raise NotImplementedError

    def temporal_link(self, data: data.FileData, extras: dict[str, Any]) -> str:
        """Return temporal download link.

        extras["ttl"] controls lifetime of the link(30 seconds by default).

        """
        raise NotImplementedError

    def one_time_link(self, data: data.FileData, extras: dict[str, Any]) -> str:
        """Return one-time download link."""
        raise NotImplementedError

    def public_link(self, data: data.FileData, extras: dict[str, Any]) -> str:
        """Return public link."""
        raise NotImplementedError


@dataclasses.dataclass()
class Settings:
    name: str = "unknown"
    override_existing: bool = False
    location_transformers: list[str] = dataclasses.field(default_factory=list)

    _required_options: ClassVar[list[str]] = []

    def __post_init__(self, **kwargs: Any):
        for attr in self._required_options:
            if not getattr(self, attr):
                raise exceptions.MissingStorageConfigurationError(self.name, attr)


class Storage:
    """Base class for storage implementation.

    Args:
        settings: storage configuration

    Example:
        ```python
        class MyStorage(Storage):
            def make_uploader(self):
                return MyUploader(self)

            def make_reader(self):
                return MyReader(self)

            def make_manager(self):
                return MyManager(self)
        ```
    """

    # do not show storage adapter
    hidden = False

    # operations that storage performs. Will be overriden by capabilities of
    # services inside constructor.
    capabilities = utils.Capability.NONE

    settings: Settings
    SettingsFactory: type[Settings] = Settings

    UploaderFactory: type[Uploader] = Uploader
    ManagerFactory: type[Manager] = Manager
    ReaderFactory: type[Reader] = Reader

    def __str__(self):
        return self.settings.name

    def __init__(self, settings: dict[str, Any], /):
        self.settings = self.configure(settings)
        self.uploader = self.make_uploader()
        self.manager = self.make_manager()
        self.reader = self.make_reader()

        self.capabilities = self.compute_capabilities()

    def make_uploader(self):
        return self.UploaderFactory(self)

    def make_manager(self):
        return self.ManagerFactory(self)

    def make_reader(self):
        return self.ReaderFactory(self)

    @classmethod
    def configure(cls, settings: dict[str, Any]) -> Any:
        try:
            return cls.SettingsFactory(**settings)
        except TypeError as err:
            raise exceptions.InvalidStorageConfigurationError(
                settings.get("name") or cls, str(err)
            ) from err

        # fields = dataclasses.fields(cls.SettingsFactory)
        # cls.SettingsFactory
        # names = {field.name for field in fields}  # initfields lost here

        # valid = {}
        # invalid = []
        # for k, v in settings.items():
        #     if k in names:
        #         valid[k] = v
        #     else:
        #         invalid.append(k)

        # cfg = cls.SettingsFactory(**valid)
        # if invalid:
        #     log.debug(
        #         "Storage %s received unknow settings: %s",
        #         cfg.name,
        #         invalid,
        #     )
        # return cfg

    def compute_capabilities(self) -> utils.Capability:
        return (
            self.uploader.capabilities
            | self.manager.capabilities
            | self.reader.capabilities
        )

    def supports(self, operation: utils.Capability) -> bool:
        return self.capabilities.can(operation)

    def prepare_location(self, location: str, /, **kwargs: Any) -> types.Location:
        for name in self.settings.location_transformers:
            if transformer := location_transformers.get(name):
                location = transformer(location, kwargs)
            else:
                raise exceptions.LocationTransformerError(name)

        return types.Location(location)

    def stream_as_upload(self, data: data.FileData, **kwargs: Any) -> Upload:
        """Make an Upload with file content."""
        stream = self.stream(data, **kwargs)
        if hasattr(stream, "read"):
            stream = cast(types.PStream, stream)
        else:
            stream = utils.IterableBytesReader(stream)

        return Upload(
            stream,
            data.location,
            data.size,
            data.content_type,
        )

    @requires_capability(utils.Capability.CREATE)
    def upload(
        self, location: types.Location, upload: Any, /, **kwargs: Any
    ) -> data.FileData:
        return self.uploader.upload(location, make_upload(upload), kwargs)

    @requires_capability(utils.Capability.MULTIPART)
    def multipart_start(
        self,
        location: types.Location,
        data: data.MultipartData,
        /,
        **kwargs: Any,
    ) -> data.MultipartData:
        return self.uploader.multipart_start(location, data, kwargs)

    @requires_capability(utils.Capability.MULTIPART)
    def multipart_refresh(
        self, data: data.MultipartData, /, **kwargs: Any
    ) -> data.MultipartData:
        return self.uploader.multipart_refresh(data, kwargs)

    @requires_capability(utils.Capability.MULTIPART)
    def multipart_update(
        self, data: data.MultipartData, /, **kwargs: Any
    ) -> data.MultipartData:
        return self.uploader.multipart_update(data, kwargs)

    @requires_capability(utils.Capability.MULTIPART)
    def multipart_complete(
        self, data: data.MultipartData, /, **kwargs: Any
    ) -> data.FileData:
        return self.uploader.multipart_complete(data, kwargs)

    @requires_capability(utils.Capability.EXISTS)
    def exists(self, data: data.FileData, /, **kwargs: Any) -> bool:
        return self.manager.exists(data, kwargs)

    @requires_capability(utils.Capability.REMOVE)
    def remove(
        self, data: data.FileData | data.MultipartData, /, **kwargs: Any
    ) -> bool:
        return self.manager.remove(data, kwargs)

    @requires_capability(utils.Capability.SCAN)
    def scan(self, **kwargs: Any) -> Iterable[str]:
        return self.manager.scan(kwargs)

    @requires_capability(utils.Capability.ANALYZE)
    def analyze(self, location: types.Location, /, **kwargs: Any) -> data.FileData:
        return self.manager.analyze(location, kwargs)

    @requires_capability(utils.Capability.STREAM)
    def stream(self, data: data.FileData, /, **kwargs: Any) -> Iterable[bytes]:
        return self.reader.stream(data, kwargs)

    def range(
        self,
        data: data.FileData,
        start: int = 0,
        end: int | None = None,
        /,
        **kwargs: Any,
    ) -> Iterable[bytes]:
        """Return byte-stream of the file content."""
        if self.supports(utils.Capability.RANGE):
            return self.reader.range(data, start, end, kwargs)

        if self.supports(utils.Capability.STREAM):
            if end is None:
                end = cast(int, float("inf"))

            end -= start
            if end <= 0:
                return

            for chunk in self.stream(data, **kwargs):
                if start > 0:
                    start -= len(chunk)
                    if start < 0:
                        chunk = chunk[start:]
                    else:
                        continue

                yield chunk[: end and None]
                end -= len(chunk)
                if end <= 0:
                    break
            return

        raise exceptions.UnsupportedOperationError("range", self)

    @requires_capability(utils.Capability.STREAM)
    def content(self, data: data.FileData, /, **kwargs: Any) -> bytes:
        return self.reader.content(data, kwargs)

    @requires_capability(utils.Capability.APPEND)
    def append(
        self,
        data: data.FileData,
        upload: Any,
        /,
        **kwargs: Any,
    ) -> data.FileData:
        return self.manager.append(data, make_upload(upload), kwargs)

    def copy(
        self,
        location: types.Location,
        data: data.FileData,
        dest_storage: Storage,
        /,
        **kwargs: Any,
    ) -> data.FileData:
        if dest_storage is self and self.supports(utils.Capability.COPY):
            return self.manager.copy(location, data, kwargs)

        if self.supports(utils.Capability.STREAM) and dest_storage.supports(
            utils.Capability.CREATE,
        ):
            return dest_storage.upload(
                location,
                self.stream_as_upload(data, **kwargs),
                **kwargs,
            )

        raise exceptions.UnsupportedOperationError("copy", self)

    def move(
        self,
        location: types.Location,
        data: data.FileData,
        dest_storage: Storage,
        /,
        **kwargs: Any,
    ) -> data.FileData:
        if dest_storage is self and self.supports(utils.Capability.MOVE):
            return self.manager.move(location, data, kwargs)

        if self.supports(
            utils.Capability.STREAM | utils.Capability.REMOVE,
        ) and dest_storage.supports(utils.Capability.CREATE):
            result = dest_storage.upload(
                location,
                self.stream_as_upload(data, **kwargs),
                **kwargs,
            )
            dest_storage.remove(data)
            return result

        raise exceptions.UnsupportedOperationError("move", self)

    def compose(
        self,
        location: types.Location,
        dest_storage: Storage,
        /,
        *files: data.FileData,
        **kwargs: Any,
    ) -> data.FileData:
        if dest_storage is self and self.supports(utils.Capability.COMPOSE):
            return self.manager.compose(location, files, kwargs)

        # REMOVE capability is required for cleaning. Alternatively, we can
        # compute expected size and check existence in advance. But this is
        # just a fallback method - use explicit implementation for efficiency.
        if self.supports(utils.Capability.STREAM) and dest_storage.supports(
            utils.Capability.CREATE | utils.Capability.APPEND | utils.Capability.REMOVE
        ):
            result = dest_storage.upload(location, make_upload(b""), **kwargs)

            # append creates file only when there are no problems. But when
            # first append succeeded with the fragment of the file added in the
            # storage, and the following append failed, this incomplete
            # fragment must be removed.
            #
            # Expected reasons of failure are:
            #
            # * one of the source fiels is missing
            # * file will go over the size limit after the following append
            try:
                for item in files:
                    result = dest_storage.append(
                        result,
                        self.stream_as_upload(item, **kwargs),
                        **kwargs,
                    )
            except (exceptions.MissingFileError, exceptions.UploadError):
                self.remove(result, **kwargs)
                raise

            return result

        raise exceptions.UnsupportedOperationError("compose", self)

    def one_time_link(self, data: data.FileData, /, **kwargs: Any) -> str | None:
        if self.supports(utils.Capability.ONE_TIME_LINK):
            return self.reader.one_time_link(data, kwargs)

    def temporal_link(self, data: data.FileData, /, **kwargs: Any) -> str | None:
        if self.supports(utils.Capability.TEMPORAL_LINK):
            return self.reader.temporal_link(data, kwargs)

    def permanent_link(self, data: data.FileData, /, **kwargs: Any) -> str | None:
        if self.supports(utils.Capability.PERMANENT_LINK):
            return self.reader.permanent_link(data, kwargs)


def make_storage(
    name: str,
    settings: dict[str, Any],
) -> Storage:
    """Initialize storage instance with specified settings.

    Storage adapter is defined by `type` key of the settings. The rest of
    settings depends on the specific adapter.

    Args:
        name: name of the storage
        settings: configuration for the storage

    Returns:
        storage instance

    Raises:
        exceptions.UnknownAdapterError: storage adapter is not registered

    Example:
        ```
        storage = make_storage("memo", {"type": "files:redis"})
        ```

    """
    adapter_type = settings.pop("type", None)
    adapter = adapters.get(adapter_type)
    if not adapter:
        raise exceptions.UnknownAdapterError(adapter_type)

    settings.setdefault("name", name)

    return adapter(settings)



from __future__ import annotations

import base64
import dataclasses
import os
import re
from typing import Any, ClassVar, Iterable

import boto3

import file_keeper as fk

RE_RANGE = re.compile(r"bytes=(?P<first_byte>\d+)-(?P<last_byte>\d+)")
HTTP_RESUME = 308


def decode(value: str) -> str:
    return base64.decodebytes(value.encode()).hex()


@dataclasses.dataclass()
class Settings(fk.Settings):
    path: str = ""

    bucket: str = ""

    key: str | None = None
    secret: str | None = None
    region: str | None = None
    endpoint: str | None = None

    client: Any = None

    _required_options: ClassVar[list[str]] = ["bucket"]

    def __post_init__(self, **kwargs: Any):
        super().__post_init__(**kwargs)

        self.path = self.path.lstrip("/")

        self.client = boto3.client(
            "s3",
            aws_access_key_id=self.key,
            aws_secret_access_key=self.secret,
            region_name=self.region,
            endpoint_url=self.endpoint,
        )


class Reader(fk.Reader):
    storage: S3Storage
    capabilities = fk.Capability.STREAM | fk.Capability.TEMPORAL_LINK

    def stream(self, data: fk.FileData, extras: dict[str, Any]) -> Iterable[bytes]:
        client = self.storage.settings.client
        filepath = os.path.join(self.storage.settings.path, data.location)

        try:
            obj: Any = client.get_object(
                Bucket=self.storage.settings.bucket, Key=filepath
            )
        except client.exceptions.NoSuchKey as err:
            raise fk.exc.MissingFileError(
                self.storage.settings.name,
                data.location,
            ) from err

        return obj["Body"]


class Uploader(fk.Uploader):
    storage: S3Storage

    capabilities = fk.Capability.CREATE | fk.Capability.MULTIPART

    def upload(
        self,
        location: fk.types.Location,
        upload: fk.Upload,
        extras: dict[str, Any],
    ) -> fk.FileData:
        filepath = os.path.join(self.storage.settings.path, location)

        client = self.storage.settings.client
        obj = client.put_object(
            Bucket=self.storage.settings.bucket, Key=filepath, Body=upload.stream
        )

        filehash = obj["ETag"].strip('"')

        return fk.FileData(
            location,
            upload.size,
            upload.content_type,
            filehash,
        )

    def multipart_start(
        self,
        location: fk.types.Location,
        data: fk.MultipartData,
        extras: dict[str, Any],
    ) -> fk.MultipartData:

        filepath = os.path.join(self.storage.settings.path, location)
        client = self.storage.settings.client
        obj = client.create_multipart_upload(
            Bucket=self.storage.settings.bucket,
            Key=filepath,
            ContentType=data.content_type,
        )

        data.location = location
        data.storage_data = dict(
            data.storage_data,
            upload_id=obj["UploadId"],
            uploaded=0,
            part_number=1,
            upload_url=self._presigned_part(filepath, obj["UploadId"], 1),
            etags={},
        )
        return data

    def _presigned_part(self, key: str, upload_id: str, part_number: int):
        return self.storage.settings.client.generate_presigned_url(
            "upload_part",
            Params={
                "Bucket": self.storage.settings.bucket,
                "Key": key,
                "UploadId": upload_id,
                "PartNumber": part_number,
            },
        )

    def multipart_update(
        self,
        data: fk.MultipartData,
        extras: dict[str, Any],
    ) -> fk.MultipartData:
        filepath = os.path.join(self.storage.settings.path, data.location)
        if "upload" in extras:
            upload: fk.Upload = extras["upload"]

            first_byte = data.storage_data["uploaded"]

            last_byte = first_byte + upload.size
            size = data.size

            if last_byte > size:
                raise fk.exc.UploadOutOfBoundError(last_byte, size)

            if upload.size < 1024 * 1024 * 5 and last_byte < size:
                raise fk.exc.ExtrasError(
                    {"upload": ["Only the final part can be smaller than 5MiB"]}
                )

            resp = self.storage.settings.client.upload_part(
                Bucket=self.storage.settings.bucket,
                Key=filepath,
                UploadId=data.storage_data["upload_id"],
                PartNumber=data.storage_data["part_number"],
                Body=upload.stream,
            )

            etag = resp["ETag"].strip('"')
            data.storage_data["uploaded"] = data.storage_data["uploaded"] + upload.size

        elif "etag" in extras:
            etag = extras["etag"].strip('"')
            data.storage_data["uploaded"] = data.storage_data["uploaded"] + extras.get(
                "uploaded", 0
            )

        else:
            raise fk.exc.ExtrasError(
                {"upload": ["Either upload or etag must be specified"]}
            )

        data.storage_data["etags"][data.storage_data["part_number"]] = etag
        data.storage_data["part_number"] = data.storage_data["part_number"] + 1

        data.storage_data["upload_url"] = self._presigned_part(
            filepath, data.storage_data["upload_id"], data.storage_data["part_number"]
        )

        return data

    def multipart_complete(
        self,
        data: fk.MultipartData,
        extras: dict[str, Any],
    ) -> fk.FileData:
        filepath = os.path.join(self.storage.settings.path, data.location)

        result = self.storage.settings.client.complete_multipart_upload(
            Bucket=self.storage.settings.bucket,
            Key=filepath,
            UploadId=data.storage_data["upload_id"],
            MultipartUpload={
                "Parts": [
                    {"PartNumber": int(num), "ETag": tag}
                    for num, tag in data.storage_data["etags"].items()
                ]
            },
        )

        obj = self.storage.settings.client.get_object(
            Bucket=self.storage.settings.bucket, Key=result["Key"]
        )

        return fk.FileData(
            fk.types.Location(
                os.path.relpath(
                    result["Key"],
                    self.storage.settings.path,
                )
            ),
            obj["ContentLength"],
            obj["ContentType"],
            obj["ETag"].strip('"'),
        )


class Manager(fk.Manager):
    storage: S3Storage

    capabilities = fk.Capability.REMOVE | fk.Capability.ANALYZE

    def remove(
        self, data: fk.FileData | fk.MultipartData, extras: dict[str, Any]
    ) -> bool:
        if isinstance(data, fk.MultipartData):
            return False

        filepath = os.path.join(str(self.storage.settings.path), data.location)
        client = self.storage.settings.client

        # TODO: check if file exists before removing to return correct status
        client.delete_object(Bucket=self.storage.settings.bucket, Key=filepath)

        return True

    def analyze(
        self, location: fk.types.Location, extras: dict[str, Any]
    ) -> fk.FileData:
        """Return all details about location."""
        filepath = os.path.join(str(self.storage.settings.path), location)
        client = self.storage.settings.client

        try:
            obj = client.get_object(Bucket=self.storage.settings.bucket, Key=filepath)
        except client.exceptions.NoSuchKey as err:
            raise fk.exc.MissingFileError(self.storage, filepath) from err

        return fk.FileData(
            location,
            size=obj["ContentLength"],
            content_type=obj["ContentType"],
            hash=obj["ETag"].strip('"'),
        )


class S3Storage(fk.Storage):
    settings: Settings  # type: ignore
    SettingsFactory = Settings
    UploaderFactory = Uploader
    ManagerFactory = Manager
    ReaderFactory = Reader



from __future__ import annotations

import dataclasses
from typing import Any, ClassVar, Iterable

from libcloud.base import DriverType, get_driver
from libcloud.common.types import LibcloudError
from libcloud.storage.base import Container, StorageDriver
from libcloud.storage.types import ContainerDoesNotExistError, ObjectDoesNotExistError

import file_keeper as fk

get_driver: Any


@dataclasses.dataclass()
class Settings(fk.Settings):
    provider: str = ""
    key: str = ""
    container_name: str = ""

    secret: str | None = None
    params: dict[str, Any] = dataclasses.field(default_factory=dict)

    driver: StorageDriver = None  # type: ignore
    container: Container = None  # type: ignore

    _required_options: ClassVar[list[str]] = ["provider", "key", "container_name"]

    def __post_init__(self, **kwargs: Any):
        super().__post_init__(**kwargs)

        try:
            make_driver = get_driver(DriverType.STORAGE, self.provider)
        except AttributeError as err:
            raise fk.exc.InvalidStorageConfigurationError(
                type(self),
                str(err),
            ) from err

        self.driver = make_driver(self.key, self.secret, **self.params)

        try:
            self.container = self.driver.get_container(self.container_name)

        except ContainerDoesNotExistError as err:
            msg = f"Container {self.container_name} does not exist"
            raise fk.exc.InvalidStorageConfigurationError(type(self), msg) from err

        except LibcloudError as err:
            raise fk.exc.InvalidStorageConfigurationError(
                type(self),
                str(err),
            ) from err


class Uploader(fk.Uploader):
    storage: LibCloudStorage
    capabilities = fk.Capability.CREATE

    def upload(
        self,
        location: fk.types.Location,
        upload: fk.Upload,
        extras: dict[str, Any],
    ) -> fk.FileData:
        result = self.storage.settings.container.upload_object_via_stream(
            iter(upload.stream),
            location,
            extra={"content_type": upload.content_type},
        )

        return fk.FileData(
            fk.types.Location(result.name),
            result.size,
            upload.content_type,
            result.hash.strip('"'),
        )


class Reader(fk.Reader):
    storage: LibCloudStorage
    capabilities = fk.Capability.STREAM | fk.Capability.TEMPORAL_LINK

    def stream(self, data: fk.FileData, extras: dict[str, Any]) -> Iterable[bytes]:
        try:
            obj = self.storage.settings.container.get_object(data.location)
        except ObjectDoesNotExistError as err:
            raise fk.exc.MissingFileError(
                self.storage,
                data.location,
            ) from err

        return obj.as_stream()


class Manager(fk.Manager):
    storage: LibCloudStorage
    capabilities = fk.Capability.SCAN | fk.Capability.REMOVE

    def scan(self, extras: dict[str, Any]) -> Iterable[str]:
        for item in self.storage.settings.container.iterate_objects():
            yield item.name

    def remove(
        self,
        data: fk.FileData | fk.MultipartData,
        extras: dict[str, Any],
    ) -> bool:
        try:
            obj = self.storage.settings.container.get_object(data.location)
        except ObjectDoesNotExistError:
            return False
        return self.storage.settings.container.delete_object(obj)


class LibCloudStorage(fk.Storage):
    settings: Settings  # type: ignore
    SettingsFactory = Settings
    UploaderFactory = Uploader
    ManagerFactory = Manager
    ReaderFactory = Reader



from __future__ import annotations

from typing import Any, Iterator, NewType, Protocol

Location = NewType("Location", str)


class PReadable(Protocol):
    def read(self, size: Any = ..., /) -> bytes: ...

class PStream(PReadable, Protocol):
    def __iter__(self) -> Iterator[bytes]: ...


class PSeekableStream(PStream):
    def tell(self) -> int: ...

    def seek(self, offset: int, whence: int = 0) -> int: ...


class PData(Protocol):
    location: Location
    size: int
    content_type: str
    hash: str
    storage_data: dict[str, Any]


__all__ = [
    "PData",
    "PStream",
    "PSeekableStream",
]



"""Base abstract functionality of the extentsion.

All classes required for specific storage implementations are defined
here. Some utilities, like `make_storage` are also added to this module instead
of `utils` to avoid import cycles.

This module relies only on types, exceptions and utils to prevent import
cycles.

"""

from __future__ import annotations

import copy
import dataclasses
from typing import Any, Generic

from typing_extensions import TypeVar

from . import types

TData = TypeVar("TData", bound=types.PData, default=Any)


@dataclasses.dataclass
class BaseData(Generic[TData]):
    location: types.Location
    size: int = 0
    content_type: str = ""
    hash: str = ""
    storage_data: dict[str, Any] = dataclasses.field(default_factory=dict)

    _plain_keys = ["location", "size", "content_type", "hash"]
    _complex_keys = ["storage_data"]

    @classmethod
    def from_dict(cls, record: dict[str, Any]):
        return cls(
            *[record[key] for key in cls._plain_keys if key in record],
            *[copy.deepcopy(record[key]) for key in cls._complex_keys if key in record],
        )

    @classmethod
    def from_object(cls, obj: TData):
        return cls(
            *[getattr(obj, key) for key in cls._plain_keys],
            *[copy.deepcopy(getattr(obj, key)) for key in cls._complex_keys],
        )

    def into_object(self, obj: TData):
        for key in self._plain_keys:
            setattr(obj, key, getattr(self, key))

        for key in self._complex_keys:
            setattr(obj, key, copy.deepcopy(getattr(self, key)))

        return obj


@dataclasses.dataclass
class FileData(BaseData[TData]):
    """Information required by storage to operate the file.

    Args:
        location: filepath, filename or any other type of unique identifier
        size: size of the file in bytes
        content_type: MIMEtype of the file
        hash: checksum of the file
        storage_data: additional details set by storage adapter

    Example:
        ```
        FileData(
            "local/path.txt",
            123,
            "text/plain",
            md5_of_content,
        )
        ```
    """

    content_type: str = "application/octet-stream"


@dataclasses.dataclass
class MultipartData(BaseData[TData]):
    """Information required by storage to operate the incomplete upload.

    Args:
        location: filepath, filename or any other type of unique identifier
        size: expected size of the file in bytes
        content_type: expected MIMEtype of the file
        hash: expected checksum of the file
        storage_data: additional details set by storage adapter

    Example:
        ```
        MultipartData(
            "local/path.txt",
            expected_size,
            expected_content_type,
            expected_hash,
        )
        ```
    """

    location: types.Location = types.Location("")



from __future__ import annotations

import copy
import dataclasses
import io
import os
from typing import Any, Iterable

import magic
import opendal

import file_keeper as fk


class FileStream:
    def __init__(self, file: opendal.File):
        self.file = file

    def __iter__(self):
        while chunk := self.file.read(io.DEFAULT_BUFFER_SIZE):
            if isinstance(chunk, memoryview):  # type: ignore
                yield chunk.tobytes()
            else:
                yield chunk


@dataclasses.dataclass()
class Settings(fk.Settings):
    params: dict[str, Any] = dataclasses.field(default_factory=dict)
    scheme: str = ""
    recursive: bool = False
    operator: opendal.Operator = None  # type: ignore

    def __post_init__(self, **kwargs: Any):
        super().__post_init__(**kwargs)

        if not self.operator:
            if not self.scheme:
                raise fk.exc.MissingStorageConfigurationError(self.name, "scheme")

            try:
                self.operator = opendal.Operator(self.scheme, **self.params)
            except opendal.exceptions.ConfigInvalid as err:  # type: ignore
                raise fk.exc.InvalidStorageConfigurationError(
                    type(self),
                    str(err),
                ) from err


class Uploader(fk.Uploader):
    storage: OpenDalStorage
    capabilities = fk.Capability.CREATE

    def upload(
        self,
        location: fk.types.Location,
        upload: fk.Upload,
        extras: dict[str, Any],
    ) -> fk.FileData:
        """Upload file to computed location.

        Schemas working with filesystem-like paths assume that location is
        relative the configured `path`. When recursive is disabled, location
        can include only filename. If `recursive` uploads allowed, location can
        be prepended with an intermediate path. This path is not sanitized and
        can lead outside the configured `path` of the storage. Consider using
        combination of `storage.prepare_location` with
        `settings.location_transformers` that sanitizes the path, like
        `safe_relative_path`.

        Raises:
            ExistingFileError: file exists and overrides are not allowed
            LocationError: unallowed usage of subdirectory

        Returns:
            New file data

        """
        op = self.storage.settings.operator
        subpath = os.path.dirname(location)

        if subpath and not (
            self.storage.settings.recursive and op.capability().create_dir
        ):
            raise fk.exc.LocationError(self.storage, subpath)

        try:
            op.stat(location)
        except opendal.exceptions.NotFound:  # type: ignore
            pass
        else:
            if not self.storage.settings.override_existing:
                raise fk.exc.ExistingFileError(self.storage, location)

        if subpath:
            # opendal reuquires `/` as a last character
            op.create_dir(subpath.rstrip("/") + "/")

        reader = upload.hashing_reader()
        with op.open(location, "wb") as dest:
            for chunk in reader:
                dest.write(chunk)

        return fk.FileData(
            location,
            reader.position,
            upload.content_type,
            reader.get_hash(),
        )


class Reader(fk.Reader):
    storage: OpenDalStorage
    capabilities = fk.Capability.STREAM

    def stream(self, data: fk.FileData, extras: dict[str, Any]) -> Iterable[bytes]:
        """Return file open in binary-read mode.

        Raises:
            MissingFileError: file does not exist

        Returns:
            File content iterator
        """
        try:
            content = self.storage.settings.operator.open(data.location, "rb")
        except opendal.exceptions.NotFound as err:  # type: ignore
            raise fk.exc.MissingFileError(self.storage, data.location) from err

        return FileStream(content)


class Manager(fk.Manager):
    storage: OpenDalStorage
    capabilities = (
        fk.Capability.REMOVE
        | fk.Capability.SCAN
        | fk.Capability.EXISTS
        | fk.Capability.COPY
        | fk.Capability.MOVE
        | fk.Capability.ANALYZE
        | fk.Capability.APPEND
    )

    def copy(
        self, location: fk.types.Location, data: fk.FileData, extras: dict[str, Any]
    ) -> fk.FileData:
        """Copy file inside the storage.

        Raises:
            ExistingFileError: file exists and overrides are not allowed
            MissingFileError: source file does not exist
        """
        op = self.storage.settings.operator

        if not self.exists(data, extras):
            raise fk.exc.MissingFileError(self.storage, data.location)

        if not self.storage.settings.override_existing and self.exists(
            fk.FileData(location), extras
        ):
            raise fk.exc.ExistingFileError(self.storage, location)

        op.copy(data.location, location)

        new_data = copy.deepcopy(data)
        new_data.location = location
        return new_data

    def move(
        self, location: fk.types.Location, data: fk.FileData, extras: dict[str, Any]
    ) -> fk.FileData:
        """Move file to a different location inside the storage.

        Raises:
            ExistingFileError: file exists and overrides are not allowed
            MissingFileError: source file does not exist
        """
        op = self.storage.settings.operator

        if not self.exists(data, extras):
            raise fk.exc.MissingFileError(self.storage, data.location)

        if not self.storage.settings.override_existing and self.exists(
            fk.FileData(location), extras
        ):
            raise fk.exc.ExistingFileError(self.storage, location)

        op.rename(data.location, location)

        new_data = copy.deepcopy(data)
        new_data.location = location
        return new_data

    def exists(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        """Check if file exists."""
        try:
            self.storage.settings.operator.stat(data.location)
        except opendal.exceptions.NotFound:  # type: ignore
            return False

        return True

    def analyze(
        self, location: fk.types.Location, extras: dict[str, Any]
    ) -> fk.FileData:
        """Check if file exists."""
        stream = self.storage.stream(fk.FileData(location))

        reader = fk.HashingReader(fk.IterableBytesReader(stream))
        content_type = magic.from_buffer(next(reader, b""), True)
        reader.exhaust()

        return fk.FileData(
            location,
            size=reader.position,
            content_type=content_type,
            hash=reader.get_hash(),
        )

    def remove(
        self,
        data: fk.FileData | fk.MultipartData,
        extras: dict[str, Any],
    ) -> bool:
        """Remove the file."""
        op = self.storage.settings.operator
        try:
            op.stat(data.location)
        except opendal.exceptions.NotFound:  # type: ignore
            return False

        op.delete(data.location)
        return True

    def scan(self, extras: dict[str, Any]) -> Iterable[str]:
        """Discover filenames in the storage."""
        for entry in self.storage.settings.operator.scan(""):
            stat = self.storage.settings.operator.stat(entry.path)
            if opendal.EntryMode.is_file(stat.mode):
                yield entry.path

    def append(
        self, data: fk.FileData, upload: fk.Upload, extras: dict[str, Any]
    ) -> fk.FileData:
        """Append content to existing file.

        If final content type is not supported by the storage, original file is
        removed.

        Raises:
            MissingFileError: file does not exist
        """
        op = self.storage.settings.operator

        if not self.exists(data, extras):
            raise fk.exc.MissingFileError(self.storage, data.location)

        op.write(data.location, upload.stream.read(), append=True)

        return self.analyze(data.location, extras)



class OpenDalStorage(fk.Storage):
    settings: Settings  # type: ignore
    SettingsFactory = Settings
    UploaderFactory = Uploader
    ManagerFactory = Manager
    ReaderFactory = Reader

    def compute_capabilities(self) -> fk.Capability:
        cluster = super().compute_capabilities()
        capabilities = self.settings.operator.capability()

        if not capabilities.stat:
            cluster = cluster.exclude(fk.Capability.EXISTS | fk.Capability.ANALYZE)

        if not capabilities.delete:
            cluster = cluster.exclude(fk.Capability.REMOVE)

        if not capabilities.list or not capabilities.stat:
            cluster = cluster.exclude(fk.Capability.SCAN)

        if not capabilities.write:
            cluster = cluster.exclude(fk.Capability.CREATE)

        if not capabilities.read:
            cluster = cluster.exclude(fk.Capability.STREAM)

        if not capabilities.rename:
            cluster = cluster.exclude(fk.Capability.MOVE)

        if not capabilities.copy:
            cluster = cluster.exclude(fk.Capability.COPY)

        if not capabilities.write_can_append:
            cluster = cluster.exclude(fk.Capability.APPEND)

        return cluster



from __future__ import annotations

import copy
import dataclasses
from io import BytesIO
from typing import IO, Any, ClassVar, Iterable, cast

import magic
import redis

import file_keeper as fk

pools = fk.Registry[redis.ConnectionPool]()


@dataclasses.dataclass
class Settings(fk.Settings):
    """Settings for Redis storage.

    Args:
        redis: existing redis connection
        redis_url: URL of the redis DB. Used only if `redis` is empty
    """

    path: str = ""
    redis: redis.Redis[bytes] = None  # type: ignore

    redis_url: dataclasses.InitVar[str] = ""

    _required_options: ClassVar[list[str]] = ["path"]

    def __post_init__(self, redis_url: str, **kwargs: Any):
        super().__post_init__(**kwargs)

        if not self.redis:
            if redis_url not in pools:
                pools.register(
                    redis_url,
                    redis.ConnectionPool.from_url(redis_url)
                    if redis_url
                    else redis.ConnectionPool(),
                )

            self.redis = redis.Redis(connection_pool=pools[redis_url])


class Uploader(fk.Uploader):
    storage: RedisStorage
    capabilities = fk.Capability.CREATE | fk.Capability.MULTIPART

    def upload(
        self,
        location: fk.types.Location,
        upload: fk.Upload,
        extras: dict[str, Any],
    ) -> fk.FileData:
        """Upload file to into location within storage path.

        Raises:
            ExistingFileError: file exists and overrides are not allowed

        Returns:
            New file data
        """
        cfg = self.storage.settings

        if not cfg.override_existing and cfg.redis.hexists(cfg.path, location):
            raise fk.exc.ExistingFileError(self.storage, location)

        reader = fk.HashingReader(upload.stream)

        content: Any = reader.read()
        cfg.redis.hset(cfg.path, location, content)

        return fk.FileData(
            location,
            reader.position,
            upload.content_type,
            reader.get_hash(),
        )

    def multipart_start(
        self,
        location: fk.types.Location,
        data: fk.MultipartData,
        extras: dict[str, Any],
    ) -> fk.MultipartData:
        """Create an empty file using `upload` method.

        Put `uploaded=0` into `data.storage_data` and copy the `location` from
        the newly created empty file.

        Returns:
            New file data
        """
        upload = fk.Upload(
            BytesIO(),
            location,
            data.size,
            data.content_type,
        )
        tmp_result = self.upload(location, upload, extras)

        data.location = tmp_result.location
        data.storage_data = dict(tmp_result.storage_data, uploaded=0)
        return data

    def multipart_refresh(
        self,
        data: fk.MultipartData,
        extras: dict[str, Any],
    ) -> fk.MultipartData:
        """Synchronize `storage_data["uploaded"]` with actual value.

        Raises:
            MissingFileError: location does not exist

        Returns:
            Updated file data
        """
        cfg = self.storage.settings

        if not cfg.redis.hexists(cfg.path, data.location):
            raise fk.exc.MissingFileError(self.storage, data.location)

        data.storage_data["uploaded"] = cfg.redis.hstrlen(cfg.path, data.location)

        return data

    def multipart_update(
        self,
        data: fk.MultipartData,
        extras: dict[str, Any],
    ) -> fk.MultipartData:
        """Add part to existing multipart upload.

        The content of upload is taken from `extras["upload"]`.

        In the end, `storage_data["uploaded"]` is set to the actial space taken
        by the storage in the system after the update.

        Raises:
            MissingFileError: file is missing
            MissingExtrasError: extra parameters are missing
            UploadOutOfBoundError: part exceeds allocated file size

        Returns:
            Updated file data
        """
        cfg = self.storage.settings

        if not cfg.redis.hexists(cfg.path, data.location):
            raise fk.exc.MissingFileError(self.storage, data.location)

        if "upload" not in extras:
            raise fk.exc.MissingExtrasError("upload")
        upload = fk.make_upload(extras["upload"])

        current = cast(bytes, cfg.redis.hget(cfg.path, data.location))
        size = len(current)

        if "uploaded" not in data.storage_data:
            data.storage_data["uploaded"] = size

        expected_size = size + upload.size
        if expected_size > data.size:
            raise fk.exc.UploadOutOfBoundError(expected_size, data.size)

        new_content: Any = current + upload.stream.read()
        cfg.redis.hset(cfg.path, data.location, new_content)

        data.storage_data["uploaded"] = expected_size
        return data

    def multipart_complete(
        self,
        data: fk.MultipartData,
        extras: dict[str, Any],
    ) -> fk.FileData:
        """Finalize the upload.

        Raises:
            MissingFileError: file does not exist
            UploadSizeMismatchError: actual and expected sizes are different
            UploadTypeMismatchError: actual and expected content types are different
            UploadHashMismatchError: actual and expected content hashes are different

        Returns:
            File data
        """
        cfg = self.storage.settings
        content = cast("bytes | None", cfg.redis.hget(cfg.path, data.location))
        if content is None:
            raise fk.exc.MissingFileError(self.storage, data.location)

        size = len(content)
        if size != data.size:
            raise fk.exc.UploadSizeMismatchError(size, data.size)

        reader = fk.HashingReader(BytesIO(content))

        content_type = magic.from_buffer(next(reader, b""), True)
        if data.content_type and content_type != data.content_type:
            raise fk.exc.UploadTypeMismatchError(
                content_type,
                data.content_type,
            )
        reader.exhaust()

        if data.hash and data.hash != reader.get_hash():
            raise fk.exc.UploadHashMismatchError(reader.get_hash(), data.hash)

        return fk.FileData(data.location, size, content_type, reader.get_hash())


class Reader(fk.Reader):
    storage: RedisStorage
    capabilities = fk.Capability.STREAM

    def stream(self, data: fk.FileData, extras: dict[str, Any]) -> IO[bytes]:
        """Return file open in binary-read mode.

        Returns:
            File content iterator
        """
        return BytesIO(self.content(data, extras))

    def content(self, data: fk.FileData, extras: dict[str, Any]) -> bytes:
        """Return content of the file.

        Raises:
            MissingFileError: file does not exist
        """
        cfg = self.storage.settings
        content = cast("bytes | None", cfg.redis.hget(cfg.path, data.location))
        if content is None:
            raise fk.exc.MissingFileError(self.storage, data.location)

        return content


class Manager(fk.Manager):
    storage: RedisStorage

    capabilities = (
        fk.Capability.COPY
        | fk.Capability.MOVE
        | fk.Capability.REMOVE
        | fk.Capability.EXISTS
        | fk.Capability.SCAN
        | fk.Capability.ANALYZE
    )

    def copy(
        self,
        location: fk.types.Location,
        data: fk.FileData,
        extras: dict[str, Any],
    ) -> fk.FileData:
        """Copy file inside the storage.

        Raises:
            ExistingFileError: file exists and overrides are not allowed
            MissingFileError: source file does not exist
        """
        cfg = self.storage.settings

        if not cfg.redis.hexists(cfg.path, data.location):
            raise fk.exc.MissingFileError(self.storage, data.location)

        if not self.storage.settings.override_existing and cfg.redis.hexists(
            cfg.path, location
        ):
            raise fk.exc.ExistingFileError(self.storage, location)

        content: Any = cfg.redis.hget(cfg.path, data.location)
        cfg.redis.hset(cfg.path, location, content)

        new_data = copy.deepcopy(data)
        new_data.location = location
        return new_data

    def move(
        self,
        location: fk.types.Location,
        data: fk.FileData,
        extras: dict[str, Any],
    ) -> fk.FileData:
        """Move file to a different location inside the storage.

        Raises:
            ExistingFileError: file exists and overrides are not allowed
            MissingFileError: source file does not exist
        """
        cfg = self.storage.settings

        if not cfg.redis.hexists(cfg.path, data.location):
            raise fk.exc.MissingFileError(self.storage, data.location)

        if not cfg.override_existing and cfg.redis.hexists(cfg.path, location):
            raise fk.exc.ExistingFileError(self.storage, location)

        content: Any = cfg.redis.hget(
            cfg.path,
            data.location,
        )
        cfg.redis.hset(cfg.path, location, content)
        cfg.redis.hdel(cfg.path, data.location)
        new_data = copy.deepcopy(data)
        new_data.location = location
        return new_data

    def exists(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        """Check if file exists."""
        cfg = self.storage.settings
        return bool(cfg.redis.hexists(cfg.path, data.location))

    def remove(
        self, data: fk.FileData | fk.MultipartData, extras: dict[str, Any]
    ) -> bool:
        """Remove the file."""
        cfg = self.storage.settings
        result = cfg.redis.hdel(cfg.path, data.location)
        return bool(result)

    def scan(self, extras: dict[str, Any]) -> Iterable[str]:
        """Discover filenames under storage path."""
        cfg = self.storage.settings
        for key in cast("Iterable[bytes]", cfg.redis.hkeys(cfg.path)):
            yield key.decode()

    def analyze(
        self, location: fk.types.Location, extras: dict[str, Any]
    ) -> fk.FileData:
        """Return all details about location.

        Raises:
            MissingFileError: file does not exist
        """
        cfg = self.storage.settings
        value: Any = cfg.redis.hget(cfg.path, location)
        if value is None:
            raise fk.exc.MissingFileError(self.storage, location)

        reader = fk.HashingReader(BytesIO(value))
        content_type = magic.from_buffer(next(reader, b""), True)
        reader.exhaust()

        return fk.FileData(
            location,
            size=reader.position,
            content_type=content_type,
            hash=reader.get_hash(),
        )


class RedisStorage(fk.Storage):
    settings: Settings  # type: ignore
    SettingsFactory = Settings

    ReaderFactory = Reader
    ManagerFactory = Manager
    UploaderFactory = Uploader



from __future__ import annotations

import copy
import dataclasses
import glob
import logging
import os
import shutil
from io import BytesIO
from typing import IO, Any, ClassVar, Iterable

import magic

import file_keeper as fk

log = logging.getLogger(__name__)


@dataclasses.dataclass()
class Settings(fk.Settings):
    """Settings for FS storage.

    Args:
        create_path: create `path` if it does not exist
        recursive: expect files inside subfolders the `path`
        path: non-empty location for storage folder
    """

    create_path: bool = False
    recursive: bool = False
    path: str = ""

    _required_options: ClassVar[list[str]] = ["path"]


class Uploader(fk.Uploader):
    storage: FsStorage
    capabilities = fk.Capability.CREATE | fk.Capability.MULTIPART

    def upload(
        self,
        location: fk.types.Location,
        upload: fk.Upload,
        extras: dict[str, Any],
    ) -> fk.FileData:
        """Upload file to computed location.

        File location is relative the configured `path`. When recursive is
        disabled, location can include only filename. If `recursive` uploads
        allowed, location can be prepended with an intermediate path. This path
        is not sanitized and can lead outside the configured `path` of the
        storage. Consider using combination of `storage.prepare_location` with
        `settings.location_transformers` that sanitizes the path, like
        `safe_relative_path`.

        Raises:
            ExistingFileError: file exists and overrides are not allowed
            LocationError: unallowed usage of subdirectory

        Returns:
            New file data

        """
        subpath = os.path.split(location)[0]

        if subpath and not self.storage.settings.recursive:
            raise fk.exc.LocationError(self.storage, subpath)

        dest = os.path.join(self.storage.settings.path, location)

        if os.path.exists(dest) and not self.storage.settings.override_existing:
            raise fk.exc.ExistingFileError(self.storage, location)

        # `recursive` is checked earlier and either subpath is empty(no
        # directories created on the next line) or `reqursive` is
        # enabled(creation of intermediate path is allowed)
        os.makedirs(os.path.dirname(dest), exist_ok=True)

        reader = upload.hashing_reader()
        with open(dest, "wb") as fd:
            for chunk in reader:
                fd.write(chunk)

        return fk.FileData(
            location,
            os.path.getsize(dest),
            upload.content_type,
            reader.get_hash(),
        )

    def multipart_start(
        self,
        location: fk.types.Location,
        data: fk.MultipartData,
        extras: dict[str, Any],
    ) -> fk.MultipartData:
        """Create an empty file using `upload` method.

        Put `uploaded=0` into `data.storage_data` and copy the `location` from
        the newly created empty file.

        Returns:
            New file data
        """
        upload = fk.Upload(
            BytesIO(),
            location,
            data.size,
            data.content_type,
        )

        tmp_result = self.upload(location, upload, extras)

        data.location = tmp_result.location
        data.storage_data = dict(tmp_result.storage_data, uploaded=0)
        return data

    def multipart_refresh(
        self,
        data: fk.MultipartData,
        extras: dict[str, Any],
    ) -> fk.MultipartData:
        """Synchronize `storage_data["uploaded"]` with actual value.

        Raises:
            MissingFileError: location does not exist

        Returns:
            Updated file data
        """
        filepath = os.path.join(
            self.storage.settings.path,
            data.location,
        )

        if not os.path.exists(filepath):
            raise fk.exc.MissingFileError(self.storage, data.location)

        data.storage_data["uploaded"] = os.path.getsize(filepath)

        return data

    def multipart_update(
        self,
        data: fk.MultipartData,
        extras: dict[str, Any],
    ) -> fk.MultipartData:
        """Add part to existing multipart upload.

        The content of upload is taken from `extras["upload"]`.

        By default, upload continues from the position specified by
        `storage_data["uploaded"]`. But if `extras["position"]` is set, it is
        used as starting point instead.

        In the end, `storage_data["uploaded"]` is set to the actial space taken
        by the file in the system after the update.

        Raises:
            UploadOutOfBoundError: part exceeds allocated file size
            MissingExtrasError: extra parameters are missing
            MissingFileError: file is missing

        Returns:
            Updated file data
        """
        filepath = os.path.join(self.storage.settings.path, data.location)

        if not os.path.exists(filepath):
            raise fk.exc.MissingFileError(self.storage, data.location)

        if "uploaded" not in data.storage_data:
            data.storage_data["uploaded"] = os.path.getsize(filepath)

        # this is the point from which upload continues. It is not used often,
        # but in specific scenario one can override previously uploaded part
        # rewinding the `position`.
        extras.setdefault("position", data.storage_data["uploaded"])

        if "upload" not in extras:
            raise fk.exc.MissingExtrasError("upload")

        upload = fk.make_upload(extras["upload"])

        # when re-uploading existing parts via explicit `position`, `uploaded`
        # can be greater than `position` + part size. For example, existing
        # content is `hello world` with size 11. One can override the first
        # word by providing content `HELLO` and position 0, resulting in
        # `position` + part size equal 5, while existing upload size remains
        # 11.
        expected_size = max(
            extras["position"] + upload.size,
            data.storage_data["uploaded"],
        )

        if expected_size > data.size:
            raise fk.exc.UploadOutOfBoundError(expected_size, data.size)

        filepath = os.path.join(self.storage.settings.path, data.location)
        with open(filepath, "rb+") as dest:
            dest.seek(extras["position"])
            for chunk in upload.stream:
                dest.write(chunk)

        data.storage_data["uploaded"] = os.path.getsize(filepath)
        return data

    def multipart_complete(
        self,
        data: fk.MultipartData,
        extras: dict[str, Any],
    ) -> fk.FileData:
        """Finalize the upload.

        Raises:
            MissingFileError: file does not exist
            UploadSizeMismatchError: actual and expected sizes are different
            UploadTypeMismatchError: actual and expected content types are different
            UploadHashMismatchError: actual and expected content hashes are different

        Returns:
            File data
        """
        filepath = os.path.join(self.storage.settings.path, data.location)
        if not os.path.exists(filepath):
            raise fk.exc.MissingFileError(self.storage, data.location)

        size = os.path.getsize(filepath)
        if size != data.size:
            raise fk.exc.UploadSizeMismatchError(size, data.size)

        with open(filepath, "rb") as src:
            reader = fk.HashingReader(src)
            content_type = magic.from_buffer(next(reader, b""), True)
            if data.content_type and content_type != data.content_type:
                raise fk.exc.UploadTypeMismatchError(
                    content_type,
                    data.content_type,
                )
            reader.exhaust()

        if data.hash and data.hash != reader.get_hash():
            raise fk.exc.UploadHashMismatchError(reader.get_hash(), data.hash)

        return fk.FileData(data.location, size, content_type, reader.get_hash())


class Reader(fk.Reader):
    storage: FsStorage
    capabilities = fk.Capability.STREAM

    def stream(self, data: fk.FileData, extras: dict[str, Any]) -> IO[bytes]:
        """Return file open in binary-read mode.

        Raises:
            MissingFileError: file does not exist

        Returns:
            File content iterator
        """
        filepath = os.path.join(self.storage.settings.path, data.location)
        if not os.path.exists(filepath):
            raise fk.exc.MissingFileError(self.storage, filepath)

        return open(filepath, "rb")  # noqa: SIM115


class Manager(fk.Manager):
    storage: FsStorage
    capabilities = (
        fk.Capability.REMOVE
        | fk.Capability.SCAN
        | fk.Capability.EXISTS
        | fk.Capability.ANALYZE
        | fk.Capability.COPY
        | fk.Capability.MOVE
        | fk.Capability.COMPOSE
        | fk.Capability.APPEND
    )

    def compose(
        self,
        location: fk.types.Location,
        datas: Iterable[fk.FileData],
        extras: dict[str, Any],
    ) -> fk.FileData:
        """Combine multipe files inside the storage into a new one.

        If final content type is not supported by the storage, the file is
        removed.

        Raises:
            ExistingFileError: file exists and overrides are not allowed
            MissingFileError: source file does not exist
        """
        dest = os.path.join(self.storage.settings.path, location)
        if os.path.exists(dest) and not self.storage.settings.override_existing:
            raise fk.exc.ExistingFileError(self.storage, location)

        sources: list[str] = []
        for data in datas:
            src = os.path.join(self.storage.settings.path, data.location)

            if not os.path.exists(src):
                raise fk.exc.MissingFileError(self.storage, src)

            sources.append(src)

        with open(dest, "wb") as to_fd:
            for src in sources:
                with open(src, "rb") as from_fd:
                    shutil.copyfileobj(from_fd, to_fd)

        return self.analyze(location, extras)

    def append(
        self,
        data: fk.FileData,
        upload: fk.Upload,
        extras: dict[str, Any],
    ) -> fk.FileData:
        """Append content to existing file.

        If final content type is not supported by the storage, original file is
        removed.

        Raises:
            MissingFileError: file does not exist
        """
        dest = os.path.join(self.storage.settings.path, data.location)
        if not os.path.exists(dest):
            raise fk.exc.MissingFileError(self.storage, dest)

        with open(dest, "ab") as fd:
            fd.write(upload.stream.read())

        return self.analyze(data.location, extras)

    def copy(
        self,
        location: fk.types.Location,
        data: fk.FileData,
        extras: dict[str, Any],
    ) -> fk.FileData:
        """Copy file inside the storage.

        Raises:
            ExistingFileError: file exists and overrides are not allowed
            MissingFileError: source file does not exist
        """
        src = os.path.join(self.storage.settings.path, data.location)
        dest = os.path.join(self.storage.settings.path, location)

        if not os.path.exists(src):
            raise fk.exc.MissingFileError(self.storage, src)

        if os.path.exists(dest) and not self.storage.settings.override_existing:
            raise fk.exc.ExistingFileError(self.storage, location)

        shutil.copy(src, dest)
        new_data = copy.deepcopy(data)
        new_data.location = location
        return new_data

    def move(
        self,
        location: fk.types.Location,
        data: fk.FileData,
        extras: dict[str, Any],
    ) -> fk.FileData:
        """Move file to a different location inside the storage.

        Raises:
            ExistingFileError: file exists and overrides are not allowed
            MissingFileError: source file does not exist
        """
        src = os.path.join(self.storage.settings.path, data.location)
        dest = os.path.join(self.storage.settings.path, location)

        if not os.path.exists(src):
            raise fk.exc.MissingFileError(self.storage, src)

        if os.path.exists(dest):
            if self.storage.settings.override_existing:
                os.remove(dest)
            else:
                raise fk.exc.ExistingFileError(self.storage, location)

        shutil.move(src, dest)
        new_data = copy.deepcopy(data)
        new_data.location = location
        return new_data

    def exists(self, data: fk.FileData, extras: dict[str, Any]) -> bool:
        """Check if file exists."""
        filepath = os.path.join(self.storage.settings.path, data.location)
        return os.path.exists(filepath)

    def remove(
        self, data: fk.FileData | fk.MultipartData, extras: dict[str, Any]
    ) -> bool:
        """Remove the file."""
        filepath = os.path.join(self.storage.settings.path, data.location)
        if not os.path.exists(filepath):
            return False

        os.remove(filepath)
        return True

    def scan(self, extras: dict[str, Any]) -> Iterable[str]:
        """Discover filenames under storage path."""
        path = self.storage.settings.path
        search_path = os.path.join(path, "**")

        for entry in glob.glob(
            search_path,
            recursive=self.storage.settings.recursive,
        ):
            if not os.path.isfile(entry):
                continue
            yield os.path.relpath(entry, path)

    def analyze(
        self, location: fk.types.Location, extras: dict[str, Any]
    ) -> fk.FileData:
        """Return all details about location.

        Raises:
            MissingFileError: file does not exist
        """
        filepath = os.path.join(self.storage.settings.path, location)
        if not os.path.exists(filepath):
            raise fk.exc.MissingFileError(self.storage, filepath)

        with open(filepath, "rb") as src:
            reader = fk.HashingReader(src)
            content_type = magic.from_buffer(next(reader, b""), True)
            reader.exhaust()

        return fk.FileData(
            location,
            size=reader.position,
            content_type=content_type,
            hash=reader.get_hash(),
        )


class FsStorage(fk.Storage):
    """Store files in local filesystem."""

    settings: Settings  # type: ignore

    SettingsFactory = Settings
    UploaderFactory = Uploader
    ReaderFactory = Reader
    ManagerFactory = Manager

    @classmethod
    def configure(cls, settings: dict[str, Any]):
        """Verify and prepare FS settings.

        Raises:
            InvalidStorageConfigurationError: incorrect configuration value
        """
        cfg: Settings = super().configure(settings)

        path = cfg.path

        if not os.path.exists(path):
            if cfg.create_path:
                os.makedirs(path)
            else:
                raise fk.exc.InvalidStorageConfigurationError(
                    cls,
                    f"path `{path}` does not exist",
                )

        return cfg



"""Internal utilities of the extension.

Do not use this module outside of the extension and do not import any other
internal module except for config, types and exceptions. Only independent tools
are stored here, to avoid import cycles.

"""

from __future__ import annotations

import abc
import enum
import hashlib
import io
import itertools
import logging
import re
from typing import Generic, Iterable, Iterator

from typing_extensions import TypeVar

from . import types

T = TypeVar("T")

log = logging.getLogger(__name__)

RE_FILESIZE = re.compile(r"^(?P<size>\d+(?:\.\d+)?)\s*(?P<unit>\w*)$")
CHUNK_SIZE = io.DEFAULT_BUFFER_SIZE
SAMPLE_SIZE = 1024 * 2
BINARY_BASE = 1024
SI_BASE = 1000

UNITS = {
    "": 1,
    "b": 1,
    "k": 10**3,
    "kb": 10**3,
    "m": 10**6,
    "mb": 10**6,
    "g": 10**9,
    "gb": 10**9,
    "t": 10**12,
    "p": 10**15,
    "tb": 10**12,
    "kib": 2**10,
    "mib": 2**20,
    "gib": 2**30,
    "tib": 2**40,
    "pib": 2**50,
}


class HashingReader:
    """IO stream wrapper that computes content hash while stream is consumed.

    Args:
        stream: iterable of bytes or file-like object
        chunk_size: max number of bytes read at once
        algorithm: hashing algorithm

    Example:
        ```
        reader = HashingReader(readable_stream)
        for chunk in reader:
            ...
        print(f"Hash: {reader.get_hash()}")
        ```
    """

    def __init__(
        self,
        stream: types.PStream,
        chunk_size: int = CHUNK_SIZE,
        algorithm: str = "md5",
    ):
        self.stream = stream
        self.chunk_size = chunk_size
        self.algorithm = algorithm
        self.hashsum = hashlib.new(algorithm)
        self.position = 0

    def __iter__(self) -> Iterator[bytes]:
        return self

    def __next__(self) -> bytes:
        chunk = self.stream.read(self.chunk_size)
        if not chunk:
            raise StopIteration

        self.position += len(chunk)
        self.hashsum.update(chunk)
        return chunk

    next = __next__

    def read(self) -> bytes:
        """Read and return all bytes from stream at once."""
        return b"".join(self)

    def get_hash(self):
        """Get current content hash as a string."""
        return self.hashsum.hexdigest()

    def exhaust(self):
        """Exhaust internal stream to compute final version of content hash.

        Note, this method does not returns data from the stream. The content
        will be irreversibly lost after method execution.
        """
        for _ in self:
            pass


class Capability(enum.Flag):
    """Enumeration of operations supported by the storage.

    Example:
        ```python
        read_and_write = Capability.STREAM | Capability.CREATE
        if storage.supports(read_and_write)
            ...
        ```
    """

    NONE = 0

    # create a file as an atomic object
    CREATE = enum.auto()
    # return file content as stream of bytes
    STREAM = enum.auto()
    # make a copy of the file inside the same storage
    COPY = enum.auto()
    # remove file from the storage
    REMOVE = enum.auto()
    # create file in 3 stages: initialize, upload(repeatable), complete
    MULTIPART = enum.auto()
    # move file to a different location inside the same storage
    MOVE = enum.auto()
    # check if file exists
    EXISTS = enum.auto()
    # iterate over all files in the storage
    SCAN = enum.auto()
    # add content to the existing file
    APPEND = enum.auto()
    # combine multiple files into a new one in the same storage
    COMPOSE = enum.auto()
    # return specific range of bytes from the file
    RANGE = enum.auto()
    # return file details from the storage, as if file was uploaded just now
    ANALYZE = enum.auto()
    # make permanent download link for file
    PERMANENT_LINK = enum.auto()
    # make expiring download link for private file
    TEMPORAL_LINK = enum.auto()
    # make one-time download link for private file
    ONE_TIME_LINK = enum.auto()

    def exclude(self, *capabilities: Capability):
        """Remove capabilities from the cluster.

        Other Args:
            capabilities: removed capabilities

        Example:
            ```python
            cluster = cluster.exclude(Capability.REMOVE)
            ```
        """
        result = Capability(self)
        for capability in capabilities:
            result = result & ~capability
        return result

    def can(self, operation: Capability) -> bool:
        return (self & operation) == operation


def parse_filesize(value: str) -> int:
    """Transform human-readable filesize into an integer.

    Args:
        value: human-readable filesize

    Raises:
        ValueError: size cannot be parsed or uses unknown units

    Example:
        ```python
        size = parse_filesize("10GiB")
        assert size == 10_737_418_240
        ```
    """
    result = RE_FILESIZE.match(value.strip())
    if not result:
        raise ValueError(value)

    size, unit = result.groups()

    multiplier = UNITS.get(unit.lower())
    if not multiplier:
        raise ValueError(value)

    return int(float(size) * multiplier)


def humanize_filesize(value: int | float, base: int = SI_BASE) -> str:
    """Transform an integer into human-readable filesize.

    Args:
        value: size in bytes

    Named args:
        base: 1000 for SI(KB, MB) or 1024 for binary(KiB, MiB)

    Raises:
        ValueError: base is not recognized

    Example:
        ```python
        size = humanize_filesize(10_737_418_240, base=1024)
        assert size == "10GiB"
        size = humanize_filesize(10_418_240, base=1024)
        assert size == "9.9MiB"
        ```

    """
    if base == SI_BASE:
        suffixes = ["B", "KB", "MB", "GB", "TB", "PB", "EB"]
    elif base == BINARY_BASE:
        suffixes = ["B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB"]
    else:
        raise ValueError(base)

    iteration = 0

    while value >= base:
        iteration += 1
        value /= base

    value = int(value * 100) / 100

    num_format = ".2f" if iteration and not value.is_integer() else ".0f"

    return f"{value:{num_format}}{suffixes[iteration]}"


class AbstractReader(Generic[T], abc.ABC):
    def __init__(self, source: T, chunk_size: int = CHUNK_SIZE):
        self.source = source
        self.chunk_size = chunk_size

    def __iter__(self):
        while chunk := self.read(self.chunk_size):
            yield chunk

    @abc.abstractmethod
    def read(self, size: int | None = None) -> bytes: ...


class IterableBytesReader(AbstractReader[Iterable[int]]):
    def __init__(self, source: Iterable[bytes], chunk_size: int = CHUNK_SIZE):
        super().__init__(itertools.chain.from_iterable(source), chunk_size)

    def read(self, size: int | None = None):
        return bytes(itertools.islice(self.source, 0, size))



__version__ = "0.0.4"

from .core import exceptions as exc
from .core import types
from .core.data import BaseData, FileData, MultipartData
from .core.registry import Registry
from .core.storage import (
    Manager,
    Reader,
    Settings,
    Storage,
    Uploader,
    adapters,
    make_storage,
)
from .core.types import Location
from .core.upload import Upload, make_upload
from .core.utils import (
    Capability,
    HashingReader,
    IterableBytesReader,
    humanize_filesize,
    parse_filesize,
)
from .ext import hookimpl  # must be the last line to avoid circular imports

__all__ = [
    "BaseData",
    "Capability",
    "Location",
    "FileData",
    "HashingReader",
    "IterableBytesReader",
    "Manager",
    "MultipartData",
    "Reader",
    "Registry",
    "Settings",
    "Storage",
    "Upload",
    "Uploader",
    "adapters",
    "exc",
    "hookimpl",
    "humanize_filesize",
    "make_storage",
    "make_upload",
    "parse_filesize",
    "types",
]



from __future__ import annotations

import base64
import dataclasses
import os
import re
from typing import Any, ClassVar, cast

import requests
from google.api_core.exceptions import Forbidden
from google.cloud.storage import Client
from google.oauth2.service_account import Credentials

import file_keeper as fk

RE_RANGE = re.compile(r"bytes=(?P<first_byte>\d+)-(?P<last_byte>\d+)")
HTTP_RESUME = 308


def decode(value: str) -> str:
    return base64.decodebytes(value.encode()).hex()


@dataclasses.dataclass()
class Settings(fk.Settings):
    path: str = ""
    bucket: str = ""
    credentials_file: str = ""
    resumable_origin: str = ""

    container: str = ""

    client: Client = None  # type: ignore
    _required_options: ClassVar[list[str]] = ["bucket", "path", "credentials_file"]

    def __post_init__(self, **kwargs: Any):
        super().__post_init__(**kwargs)
        self.path = self.path.lstrip("/")

        credentials = None
        if self.credentials_file:
            try:
                credentials = Credentials.from_service_account_file(
                    self.credentials_file
                )
            except OSError as err:
                raise fk.exc.InvalidStorageConfigurationError(
                    type(self),
                    f"file `{self.credentials_file}` does not exist",
                ) from err

        self.client = Client(credentials=credentials)


class Uploader(fk.Uploader):
    storage: GoogleCloudStorage

    capabilities = fk.Capability.CREATE | fk.Capability.MULTIPART

    def upload(
        self,
        location: fk.types.Location,
        upload: fk.Upload,
        extras: dict[str, Any],
    ) -> fk.FileData:
        filepath = os.path.join(self.storage.settings.path, location)

        client = self.storage.settings.client
        blob = client.bucket(self.storage.settings.bucket).blob(filepath)

        blob.upload_from_file(upload.stream)
        filehash = decode(blob.md5_hash)
        return fk.FileData(
            location,
            blob.size or upload.size,
            upload.content_type,
            filehash,
        )

    def multipart_start(
        self,
        location: fk.types.Location,
        data: fk.MultipartData,
        extras: dict[str, Any],
    ) -> fk.MultipartData:
        filepath = os.path.join(self.storage.settings.path, location)

        client = self.storage.settings.client
        blob = client.bucket(self.storage.settings.bucket).blob(filepath)

        url = cast(
            str,
            blob.create_resumable_upload_session(
                size=data.size,
                origin=self.storage.settings.resumable_origin,
            ),
        )

        if not url:
            msg = "Cannot initialize session URL"
            raise fk.exc.UploadError(msg)

        data.location = location
        data.storage_data = dict(
            data.storage_data,
            session_url=url,
            uploaded=0,
        )
        return data

    def multipart_update(
        self,
        data: fk.MultipartData,
        extras: dict[str, Any],
    ) -> fk.MultipartData:
        if "upload" in extras:
            upload: fk.Upload = extras["upload"]

            first_byte = extras.get("position", data.storage_data["uploaded"])
            last_byte = first_byte + upload.size - 1
            size = data.size

            if last_byte >= size:
                raise fk.exc.UploadOutOfBoundError(last_byte, size)

            if upload.size < 256 * 1024 and last_byte < size - 1:
                raise fk.exc.ExtrasError(
                    {"upload": ["Only the final part can be smaller than 256KiB"]},
                )

            resp = requests.put(
                data.storage_data["session_url"],
                data=upload.stream.read(),
                headers={
                    "content-range": f"bytes {first_byte}-{last_byte}/{size}",
                },
                timeout=10,
            )

            if not resp.ok:
                raise fk.exc.ExtrasError({"upload": [resp.text]})

            if "range" not in resp.headers:
                data.storage_data["uploaded"] = data.size
                data.storage_data["result"] = resp.json()
                return data

            range_match = RE_RANGE.match(resp.headers["range"])
            if not range_match:
                raise fk.exc.ExtrasError(
                    {"upload": ["Invalid response from Google Cloud"]},
                )
            data.storage_data["uploaded"] = int(range_match.group("last_byte")) + 1

        elif "uploaded" in extras:
            data.storage_data["uploaded"] = extras["uploaded"]

        else:
            raise fk.exc.ExtrasError(
                {"upload": ["Either upload or uploaded must be specified"]},
            )

        return data

    def multipart_refresh(
        self,
        data: fk.MultipartData,
        extras: dict[str, Any],
    ) -> fk.MultipartData:
        resp = requests.put(
            data.storage_data["session_url"],
            headers={
                "content-range": f"bytes */{data.size}",
                "content-length": "0",
            },
            timeout=10,
        )
        if not resp.ok:
            raise fk.exc.ExtrasError({"session_url": [resp.text]})

        if resp.status_code == HTTP_RESUME:
            if "range" in resp.headers:
                range_match = RE_RANGE.match(resp.headers["range"])
                if not range_match:
                    raise fk.exc.ExtrasError(
                        {
                            "session_url": [
                                "Invalid response from Google Cloud:"
                                + " missing range header",
                            ],
                        },
                    )
                data.storage_data["uploaded"] = int(range_match.group("last_byte")) + 1
            else:
                data.storage_data["uploaded"] = 0
        elif resp.status_code in [200, 201]:
            data.storage_data["uploaded"] = data.size
            data.storage_data["result"] = resp.json()

        else:
            raise fk.exc.ExtrasError(
                {
                    "session_url": [
                        "Invalid response from Google Cloud:"
                        + f" unexpected status {resp.status_code}",
                    ],
                },
            )

        return data

    def multipart_complete(
        self,
        data: fk.MultipartData,
        extras: dict[str, Any],
    ) -> fk.FileData:
        data = self.multipart_refresh(data, extras)
        if data.storage_data["uploaded"] != data.size:
            raise fk.exc.UploadSizeMismatchError(
                data.storage_data["uploaded"],
                data.size,
            )

        filehash = decode(data.storage_data["result"]["md5Hash"])
        if data.hash and data.hash != filehash:
            raise fk.exc.UploadHashMismatchError(filehash, data.hash)

        content_type = data.storage_data["result"]["contentType"]
        if data.content_type and content_type != data.content_type:
            raise fk.exc.UploadTypeMismatchError(content_type, data.content_type)

        return fk.FileData(
            fk.types.Location(
                os.path.relpath(
                    data.storage_data["result"]["name"],
                    self.storage.settings.path,
                )
            ),
            data.size,
            content_type,
            filehash,
        )


class Manager(fk.Manager):
    storage: GoogleCloudStorage
    capabilities = fk.Capability.REMOVE

    def remove(
        self, data: fk.FileData | fk.MultipartData, extras: dict[str, Any]
    ) -> bool:
        if isinstance(data, fk.MultipartData):
            return False

        filepath = os.path.join(str(self.storage.settings.path), data.location)
        client: Client = self.storage.settings.client
        blob = client.bucket(self.storage.settings.bucket).blob(filepath)

        try:
            exists = blob.exists()
        except Forbidden as err:
            raise fk.exc.PermissionError(
                self,
                "exists",
                str(err),
            ) from err

        if exists:
            try:
                blob.delete()
            except Forbidden as err:
                raise fk.exc.PermissionError(
                    self,
                    "remove",
                    str(err),
                ) from err
            return True
        return False


class GoogleCloudStorage(fk.Storage):
    settings: Settings  # type: ignore
    SettingsFactory = Settings
    UploaderFactory = Uploader
    ManagerFactory = Manager
    # ReaderFactory = Reader



from __future__ import annotations

import dataclasses
from io import BufferedReader, BytesIO
from typing import Any, Callable, cast

import magic
from typing_extensions import TypeAlias

from . import registry, types, utils

SAMPLE_SIZE = 1024 * 2

UploadFactory: TypeAlias = Callable[
    [Any], "Upload | BytesIO | BufferedReader | bytes | bytearray | None"
]

upload_factories = registry.Registry[UploadFactory, type]()

@dataclasses.dataclass
class Upload:
    """Standard upload details.

    Args:
        stream: iterable of bytes or file-like object
        filename: name of the file
        size: size of the file in bytes
        content_type: MIMEtype of the file

    Example:
        ```
        Upload(
            BytesIO(b"hello world"),
            "file.txt",
            11,
            "text/plain",
        )
        ```
    """

    stream: types.PStream
    filename: str
    size: int
    content_type: str

    @property
    def seekable_stream(self) -> types.PSeekableStream | None:
        """Return stream that can be rewinded after reading.

        If internal stream does not support file-like `seek`, nothing is
        returned from this property.

        Use this property if you want to read the file ahead, to get CSV column
        names, list of files inside ZIP, EXIF metadata. If you get `None` from
        it, stream does not support seeking and you won't be able to rewind
        cursor to the beginning of the file after reading something.

        Example:
            ```python
            upload = make_upload(...)
            if fd := upload.seekable_stream():
                # read fragment of the file
                chunk = fd.read(1024)
                # move cursor to the end of the stream
                fd.seek(0, 2)
                # position of the cursor is the same as number of bytes in stream
                size = fd.tell()
                # move cursor back, because you don't want to accidentally loose
                # any bites from the beginning of stream when uploader reads from it
                fd.seek(0)
            ```

        Returns:
            file-like stream or nothing

        """
        if hasattr(self.stream, "tell") and hasattr(self.stream, "seek"):
            return cast(types.PSeekableStream, self.stream)

        return None

    def hashing_reader(self, **kwargs: Any) -> utils.HashingReader:
        return utils.HashingReader(self.stream, **kwargs)


def make_upload(value: Any) -> Upload:
    """Convert value into Upload object.

    Use this function for simple and reliable initialization of Upload
    object. Avoid creating Upload manually, unless you are 100% sure you can
    provide correct MIMEtype, size and stream.

    Args:
        value: content of the file

    Raises:
        TypeError: content has unsupported type

    Returns:
        upload object with specified content

    Example:
        ```python
        storage.upload("file.txt", make_upload(b"hello world"))
        ```

    """
    if isinstance(value, Upload):
        return value

    initial_type: type = type(value)

    fallback_factory = None
    for t in upload_factories:
        if initial_type is t:
            transformed_value = upload_factories[t](value)
            if transformed_value is not None:
                value = transformed_value
                break

        if not fallback_factory and issubclass(initial_type, t):
            fallback_factory = upload_factories[t]

    else:
        if fallback_factory:
            value = fallback_factory(value)

    # ideal situation: factory produced the Upload object
    if isinstance(value, Upload):
        return value

    if isinstance(value, (bytes, bytearray)):
        value = BytesIO(value)

    # convenient situation: factory produced binary buffer and we know how to
    # transform it into an Upload. Factories will choose this option to avoid
    # repeating mimetype detection logic
    if isinstance(value, (BytesIO, BufferedReader)):
        mime = magic.from_buffer(value.read(SAMPLE_SIZE), True)
        _ = value.seek(0, 2)
        size = value.tell()
        _ = value.seek(0)

        return Upload(value, getattr(value, "name", ""), size, mime)

    raise TypeError(type(value))



from .fs import FsStorage

try:
    from .redis import RedisStorage
except ImportError:
    RedisStorage = None

try:
    from .opendal import OpenDalStorage
except ImportError:
    OpenDalStorage = None

try:
    from .libcloud import LibCloudStorage
except ImportError:
    LibCloudStorage = None

try:
    from .gcs import GoogleCloudStorage
except ImportError:
    GoogleCloudStorage = None

try:
    from .s3 import S3Storage
except ImportError:
    S3Storage = None

try:
    from .filebin import FilebinStorage
except ImportError:
    FilebinStorage = None

try:
    from .sqlalchemy import SqlAlchemyStorage
except ImportError:
    SqlAlchemyStorage = None

__all__ = [
    "FsStorage",
    "RedisStorage",
    "OpenDalStorage",
    "LibCloudStorage",
    "FilebinStorage",
    "GoogleCloudStorage",
    "S3Storage",
    "SqlAlchemyStorage"
]
